@@ -26,61 +26,14 @@ setClass("XVector",
 
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### Getters.
+### Getters
 ###
 
 setMethod("length", "XVector", function(x) x@length)
 
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### Combining.
-###
-
-### Should work as an endomorphism (e.g. will return a DNAString instance if
-### 'x' is a DNAString instance).
-setMethod("c", "XVector",
-    function(x, ..., recursive=FALSE)
-    {
-        if (!identical(recursive, FALSE))
-            stop("\"c\" method for XVector objects ",
-                 "does not support the 'recursive' argument")
-        if (missing(x)) {
-            args <- unname(list(...))
-            x <- args[[1L]]
-        } else {
-            args <- unname(list(x, ...))
-        }
-        if (length(args) == 1L)
-            return(x)
-        arg_is_null <- sapply(args, is.null)
-        ## Remove NULL elements by setting them to NULL!
-        if (any(arg_is_null))
-            args[arg_is_null] <- NULL
-        if (!all(sapply(args, is, class(x))))
-            stop("all arguments in '...' must be ",
-                 class(x), " objects (or NULLs)")
-        ans_length <- sum(sapply(args, length))
-        ans_shared <- SharedVector(class(x@shared), length=ans_length)
-        dest_offset <- 0L
-        for (arg in args) {
-            width <- length(arg)
-            if (width == 0L)  # will be TRUE on NULLs too...
-                next
-            ## ... so from here 'arg' is guaranteed to be an XVector object.
-            src <- arg@shared
-            src_start <- arg@offset + 1L
-            SharedVector.mcopy(ans_shared, dest_offset, src, src_start, width)
-            dest_offset <- dest_offset + width
-        }
-        ans <- new2(class(x), length=ans_length, check=FALSE)
-        ans@shared <- ans_shared
-        ans
-    }
-)
-
-
-### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### Subsetting.
+### Subsetting
 ###
 
 setMethod("extractROWS", "XVector",
@@ -136,7 +89,7 @@ setReplaceMethod("subseq", "XVector",
 
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### Coercion.
+### Coercion
 ###
 
 ### Works as long as as.integer() works on 'x'.
@@ -148,8 +101,9 @@ setAs("XVector", "Rle", function(from) {
   Rle(as.vector(from))
 })
 
+
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### The "show" method.
+### The "show" method
 ###
 
 setMethod("show", "XVector",
@@ -164,7 +118,59 @@ setMethod("show", "XVector",
 
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### Equality.
+### Concatenation
+###
+
+.concatenate_XVector_objects <-
+    function(x, objects=list(), use.names=TRUE, ignore.mcols=FALSE, check=TRUE)
+{
+    objects <- S4Vectors:::prepare_objects_to_concatenate(x, objects)
+    all_objects <- c(list(x), objects)
+
+    ans_len <- suppressWarnings(sum(lengths(all_objects)))
+    if (is.na(ans_len))
+        stop("too many vector elements to concatenate")
+
+    ## 1. Take care of the parallel slots
+
+    ## Call method for Vector objects to concatenate all the parallel
+    ## slots (only "elementMetadata" in the case of XVector) and stick them
+    ## into 'ans'. Note that the resulting 'ans' can be an invalid object
+    ## ball method for Vector objects to concatenate all the parallel
+    ## slots (only "elementMetadata" in the case of IPos) and stick them
+    ## into 'ans'. Note that the resulting 'ans' can be an invalid object
+    ## because its "elementMetadata" slot can be longer (i.e. have more rows)
+    ## than 'ans' itself so we use 'check=FALSE' to skip validation.
+    ans <- callNextMethod(x, objects, use.names=use.names,
+                                      ignore.mcols=ignore.mcols,
+                                      check=FALSE)
+
+    ## 2. Take care of the non-parallel slots
+
+    ans_shared <- SharedVector(class(x@shared), length=ans_len)
+    dest_offset <- 0L
+    for (object in all_objects) {
+        object_len <- length(object)
+        if (object_len == 0L)  # would be TRUE on NULLs too...
+            next
+        ## From here 'object' is guaranteed to be an XVector object.
+        src <- object@shared
+        src_start <- object@offset + 1L
+        SharedVector.mcopy(ans_shared, dest_offset, src, src_start, object_len)
+        dest_offset <- dest_offset + object_len
+    }
+
+    BiocGenerics:::replaceSlots(ans, shared=ans_shared,
+                                     offset=0L,
+                                     length=ans_len,
+                                     check=check)
+}
+
+setMethod("concatenateObjects", "XVector", .concatenate_XVector_objects)
+
+
+### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+### Equality
 ###
 
 .XVector.equal <- function(x, y)
