@@ -8,6 +8,7 @@
 ### current S4 implementation is *very* inefficient at.
 ###
 
+
 setClass("GroupedIRanges",
     contains="IRanges",
     representation(
@@ -26,21 +27,31 @@ setClass("XVectorList",
 
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### GroupedIRanges methods.
+### parallelSlotNames()
 ###
 
-### Ugly workaround a serious callNextMethod inefficiency reported here:
-###   https://bugs.r-project.org/bugzilla/show_bug.cgi?id=16974
-.GroupedIRanges_parallelSlotNames <-
-    c("group", parallelSlotNames(new("IRanges")))
-
 ### Combine the new parallel slots with those of the parent class. Make sure
 ### to put the new parallel slots *first*.
+
+### Ugly workaround a serious callNextMethod inefficiency reported here:
+###   https://bugs.r-project.org/bugzilla/show_bug.cgi?id=16974
+.parallelSlotNames_of_GroupedIRanges <-
+    c("group", parallelSlotNames(new("IRanges")))
+
 setMethod("parallelSlotNames", "GroupedIRanges",
     #function(x) c("group", callNextMethod())
-    function(x) .GroupedIRanges_parallelSlotNames
+    function(x) .parallelSlotNames_of_GroupedIRanges
 )
 
+setMethod("parallelSlotNames", "XVectorList",
+    function(x) c("ranges", callNextMethod())
+)
+
+
+### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+### GroupedIRanges methods
+###
+
 .valid.GroupedIRanges <- function(x)
 {
     if (length(x@group) != length(x))
@@ -59,29 +70,11 @@ setMethod("show", "GroupedIRanges",
     function(object) show(as.data.frame(object))
 )
 
-setMethod("c", "GroupedIRanges",
-    function(x, ..., recursive=FALSE)
-    {
-        if (!identical(recursive, FALSE))
-            stop("\"c\" method for GroupedIRanges objects ",
-                 "does not support the 'recursive' argument")
-        old_val <- S4Vectors:::disableValidity()
-        on.exit(S4Vectors:::disableValidity(old_val))
-        S4Vectors:::disableValidity(TRUE)
-        ans <- callNextMethod(x, ..., recursive=FALSE)
-        ans@group <-
-            do.call(c, lapply(unname(list(x, ...)), function(arg) arg@group))
-        ans
-    }
-)
-
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### XVectorList accessor-like methods.
+### XVectorList accessors
 ###
 
-setMethod("length", "XVectorList", function(x) length(x@ranges))
-
 setMethod("width", "XVectorList", function(x) width(x@ranges))
 
 setMethod("elementNROWS", "XVectorList", function(x) width(x))
@@ -99,7 +92,7 @@ setReplaceMethod("names", "XVectorList",
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### 2 internal bookkeeping functions to keep the XVectorList "pool" slot
-### clean and tidy.
+### clean and tidy
 ###
 
 ### Used in "extractROWS" method for XVectorList objects.
@@ -135,7 +128,7 @@ setReplaceMethod("names", "XVectorList",
 
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### XVectorList constructors.
+### XVectorList constructors
 ###
 
 new_XVectorList_from_list_of_XVector <- function(classname, x)
@@ -189,14 +182,14 @@ XVectorList <- function(classname, length=0L)
 
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### Going from XVector to XVectorList with extractList() and family.
+### Going from XVector to XVectorList with extractList() and family
 ###
 
 setMethod("relistToClass", "XVector",
     function(x) paste0(class(x), "List")
 )
 
-### Takes one XVector object ('x') and a Ranges object ('i') defining
+### Takes one XVector object ('x') and a IntegerRanges object ('i') defining
 ### 1-based ranges on 'x' (conceptually equivalent to defining views on
 ### subject 'x').
 .unsafe.extractList <- function(x, i)
@@ -229,7 +222,7 @@ setMethod("relist", c("XVector", "PartitioningByEnd"),
 )
 
 ### Does not copy the sequence data!
-setMethod("extractList", c("XVector", "Ranges"),
+setMethod("extractList", c("XVector", "IntegerRanges"),
     function(x, i)
     {
         if (length(i) != 0L
@@ -241,44 +234,31 @@ setMethod("extractList", c("XVector", "Ranges"),
 
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### XVectorList subsetting.
+### XVectorList subsetting
 ###
 
-### TODO: Make this a "getListElement" method for XVectorList objects.
-XVectorList.getElement <- function(x, i)
+.getListElement_XVectorList <- function(x, i, exact=TRUE)
 {
+    i2 <- normalizeDoubleBracketSubscript(i, x, exact=exact,
+                                          allow.NA=TRUE,
+                                          allow.nomatch=TRUE)
+    if (is.na(i2))
+        return(NULL)
     ans_class <- elementType(x)
-    ans_shared <- x@pool[[x@ranges@group[i]]]
-    ans_offset <- x@ranges@start[i] - 1L
-    ans_length <- x@ranges@width[i]
-    ans <- new2(ans_class,
-                shared=ans_shared,
-                offset=ans_offset,
-                length=ans_length,
-                check=FALSE)
-    return(ans)
+    ans_shared <- x@pool[[x@ranges@group[i2]]]
+    ans_offset <- x@ranges@start[i2] - 1L
+    ans_length <- x@ranges@width[i2]
+    new2(ans_class, shared=ans_shared,
+                    offset=ans_offset,
+                    length=ans_length,
+                    check=FALSE)
 }
 
-setMethod("[[", "XVectorList",
-    function(x, i, j, ..., exact=TRUE)
-    {
-        i <- normalizeDoubleBracketSubscript(i, x)
-        XVectorList.getElement(x, i)
-    }
-)
+setMethod("getListElement", "XVectorList", .getListElement_XVectorList)
 
+### Drop unused pool elements.
 setMethod("extractROWS", "XVectorList",
-    function(x, i)
-    {
-        i <- normalizeSingleBracketSubscript(i, x, as.NSBS=TRUE)
-        ans_ranges <- extractROWS(x@ranges, i)
-        ans_mcols <- extractROWS(x@elementMetadata, i)
-        ans <- BiocGenerics:::replaceSlots(x, ranges=ans_ranges,
-                                              mcols=ans_mcols,
-                                              check=FALSE)
-        ## Drop unused pool elements.
-        .dropUnusedPoolElts(ans)
-    }
+    function(x, i) .dropUnusedPoolElts(callNextMethod())
 )
 
 
@@ -293,144 +273,52 @@ setMethod("subseq", "XVectorList",
 
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### Combining.
-###
-### The "c" method for XVectorList objects is implemented to behave like an
-### endomorphism i.e. to return an object of the same class as 'x'. In
-### particular 'c(x)' returns 'x' and not 'as(x, "XVectorList")'.
-### It's easy to implement specific "c" methods for XVectorList subclasses.
-### Typically they just need to do something like:
-###
-###     old_val <- S4Vectors:::disableValidity()
-###     on.exit(S4Vectors:::disableValidity(old_val))
-###     S4Vectors:::disableValidity(TRUE)
-###     ans <- callNextMethod(x, ..., recursive=FALSE)
-###     ...
-###
-### and to take care of the additional slots (aka the subclass-specific
-### slots). If there aren't any additional slots (e.g. XRawList), or if the
-### additional slots don't need to be modified, then no need to implement a
-### specific method at all.
+### Concatenation
 ###
 
-### 'Class' must be the name of a concrete subclass that extends XVectorList.
-### Returns an instance of class 'Class'.
-unlist_list_of_XVectorList <- function(Class, x,
-                                       use.names=TRUE, ignore.mcols=FALSE)
+.concatenate_XVectorList_objects <-
+    function(x, objects=list(), use.names=TRUE, ignore.mcols=FALSE, check=TRUE)
 {
-    if (!isSingleString(Class))
-        stop("'Class' must be a single character string")
-    if (!extends(Class, "XVectorList"))
-        stop("'Class' must be the name of a class that extends XVectorList")
-    if (!is.list(x))
-        stop("'x' must be a list")
-    if (!isTRUEorFALSE(use.names))
-        stop("'use.names' must be TRUE or FALSE")
-    ### TODO: Support 'use.names=TRUE'.
-    if (use.names)
-        stop("'use.names=TRUE' is not supported yet")
-    if (!isTRUEorFALSE(ignore.mcols))
-        stop("'ignore.mcols' must be TRUE or FALSE")
+    objects <- S4Vectors:::prepare_objects_to_concatenate(x, objects)
+    all_objects <- c(list(x), objects)
 
-    ## TODO: Implement (in C) fast elementIsNull(x), that does
-    ## 'sapply(x, is.null, USE.NAMES=FALSE)' on list 'x', and use it here.
-    null_idx <- which(sapply(x, is.null, USE.NAMES=FALSE))
-    if (length(null_idx) != 0L)
-        x <- x[-null_idx]
-    if (length(x) == 0L)
-        return(new(Class))
-    ## TODO: Implement (in C) fast elementIs(x, class), that does
-    ## 'sapply(x, is, class, USE.NAMES=FALSE)' on list 'x', and use it here.
-    ## 'elementIs(x, "NULL")' should work and be equivalent to
-    ## 'elementIsNull(x)'.
-    if (!all(sapply(x, is, Class, USE.NAMES=FALSE)))
-        stop("all elements in 'x' must be ", Class, " objects (or NULLs)")
-    x_names <- names(x)
-    names(x) <- NULL  # so lapply(x, ...) below returns an unnamed list
+    ## Call method for Vector objects to concatenate all the parallel slots
+    ## (i.e. "ranges" and "elementMetadata" in the case of XVectorList) and
+    ## stick them into 'ans'. Note that the resulting 'ans' can be an invalid
+    ## object so we use 'check=FALSE' to skip validation.
+    ans <- callNextMethod(x, objects, use.names=use.names,
+                                      ignore.mcols=ignore.mcols,
+                                      check=FALSE)
 
-    ## Combine "pool" slots.
-    pool_slots <- lapply(x, function(xi) xi@pool)
-    ## TODO: Implement unlist_list_of_SharedRaw_Pool() and use it here.
-    ans_pool <- do.call(c, pool_slots)
+    ## 2. Take care of the non-parallel slots
 
-    ## Combine "ranges" slots.
-    ranges_slots <- lapply(x, function(xi) xi@ranges)
-    ## TODO: Implement unlist_list_of_GroupedRanges() (that takes an 'offsets'
-    ## arg) and use it here.
-    ans_ranges <- do.call(c, ranges_slots)
+    ## Concatenate the "pool" slots.
+    pool_list <- lapply(all_objects, slot, "pool")
+    ans_pool <- do.call(c, pool_list)
 
-    breakpoints <- cumsum(elementNROWS(pool_slots))
+    ## 3. Fix parallel slot "ranges"
+
+    ans_ranges <- ans@ranges
+    breakpoints <- cumsum(lengths(pool_list))
     offsets <- c(0L, breakpoints[-length(breakpoints)])
-    offsets <- rep.int(offsets, elementNROWS(ranges_slots))
+    offsets <- rep.int(offsets, lengths(all_objects))
     ans_ranges@group <- ans_ranges@group + offsets
+    if (!(use.names || is.null(names(ans_ranges))))
+        names(ans_ranges) <- NULL
 
-    ## Combine "mcols" slots.
-    ans_mcols <- do.call(S4Vectors:::rbind_mcols, x)
-
-    ## Make 'ans' and return it.
-    ans <- new(Class, pool=ans_pool,
-                      ranges=ans_ranges,
-                      elementMetadata=ans_mcols)
+    ans <- BiocGenerics:::replaceSlots(ans, pool=ans_pool,
+                                            ranges=ans_ranges,
+                                            check=check)
     .dropDuplicatedPoolElts(ans)
 }
 
-setMethod("c", "XVectorList",
-    function(x, ..., ignore.mcols=FALSE, recursive=FALSE)
-    {
-        if (!identical(recursive, FALSE))
-            stop("\"c\" method for XVectorList objects ",
-                 "does not support the 'recursive' argument")
-        if (missing(x)) {
-            args <- list(...)
-            x <- args[[1L]]
-        } else {
-            args <- list(x, ...)
-        }
-        unlist_list_of_XVectorList(class(x), args,
-                                   use.names=FALSE, ignore.mcols=ignore.mcols)
-    }
+setMethod("concatenateObjects", "XVectorList",
+    .concatenate_XVectorList_objects
 )
 
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### Replacement methods.
-###
-
-setReplaceMethod("[[", "XVectorList",
-    function(x, i, j, ..., value)
-    {
-        i <- normalizeDoubleBracketSubscript(i, x)
-        if (!is(value, elementType(x)))
-            stop("supplied replacement value must be a ",
-                 elementType(x), " object")
-        x[i] <- as(value, class(x))
-        x
-    }
-)
-
-
-### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### Looping methods.
-###
-
-setMethod("endoapply", "XVectorList",
-    function(X, FUN, ...)
-    {
-        Xconstructor <- get(class(X))
-        ## If there is no constructor for 'class(X)' or if the constructor
-        ## doesn't work on a list (here we try on an empty list), then we
-        ## call the default method i.e. the method for List objects. This will
-        ## be much slower but still better than failing.
-        if (!is.function(Xconstructor) ||
-            inherits(try(Xconstructor(list()), silent=TRUE), "try-error"))
-            return(callNextMethod())
-        Xconstructor(lapply(X, FUN, ...))
-    }
-)
-
-
-### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### Show method for data column.
+### Show method for data column
 ###
 
 setMethod("showAsCell", "XVectorList", function(object) as.character(object))
@@ -442,7 +330,7 @@ setMethod("showAsCell", "XVectorList", function(object) as.character(object))
 ### Not intended for the end user.
 ###
 ### 'f' must be a factor with number of levels equal to 'length(x)' and
-### length equal to 'sum(elementNROWS(x))'. 
+### length equal to 'sum(lengths(x))'.
 unsplit_list_of_XVectorList <- function(classname, x, f)
 {
     ans <- XVectorList(classname, length(f))
